"""
Fondat AWS client module.
"""

import botocore.config
import fondat.error
import logging

from aiobotocore.client import AioBaseClient
from aiobotocore.session import AioSession
from botocore.exceptions import ClientError
from contextlib import asynccontextmanager, contextmanager
from fondat.data import datacls
from typing import Annotated


_logger = logging.getLogger(__name__)


@datacls
class Config:
    profile: Annotated[str | None, "name of the profile to use"]
    region_name: Annotated[str | None, "name of the region to connect to"]
    verify: Annotated[bool | str, "verify TLS certificates or path to CA cert bundle"] = True
    endpoint_url: Annotated[str | None, "URL to use for constructed client"]
    aws_access_key_id: Annotated[str | None, "access key ID"]
    aws_secret_access_key: Annotated[str | None, "secret access key"]
    aws_session_token: Annotated[str | None, "session token"]
    config: Annotated[botocore.config.Config | None, "advanced configuration options"]


@asynccontextmanager
async def create_client(
    service_name: str,
    *,
    api_version: str | None = None,
    config: Config | None = None,
) -> AioBaseClient:
    """
    Create an aiobotocore client.

    Parameters:
    • service_name: the name of the service to access
    • api_version: the API version to use  [latest]
    • config: session and client configuration

    If no configuration is provided, or configuration parameters are omitted, then
    aiobotocore will revert to searching environment variables and files in ~/.aws/.
    """

    session = AioSession(profile=config.profile if config else None)
    async with session.create_client(
        service_name=service_name,
        region_name=config.region_name if config else None,
        api_version=api_version,
        use_ssl=True,
        verify=config.verify if config else True,
        endpoint_url=config.endpoint.url if config else None,
        aws_access_key_id=config.aws_access_key_id if config else None,
        aws_secret_access_key=config.aws_secret_access_key if config else None,
        aws_session_token=config.aws_session_token if config else None,
        config=config.config if config else None,
    ) as client:
        yield client


@contextmanager
def wrap_client_error():
    """Catch any raised ClientError and reraise as a Fondat resource error."""
    try:
        yield
    except ClientError as ce:
        if ce.response["Error"]["Code"] == "ResourceNotFoundException":
            status = 404
        else:
            status = ce.response["ResponseMetadata"]["HTTPStatusCode"]
        raise fondat.error.errors[status] from ce



"""Fondat module for Amazon Simple Storage Service (S3)."""

import fondat.aws.client
import fondat.codec
import fondat.pagination
import logging

from collections.abc import Iterable
from contextlib import asynccontextmanager
from fondat.aws.client import Config
from fondat.codec import Binary, String
from fondat.error import InternalServerError, NotFoundError
from fondat.resource import operation, resource
from fondat.security import Policy
from fondat.validation import validate_arguments
from typing import Any
from urllib.parse import quote


_logger = logging.getLogger(__name__)


@validate_arguments
def bucket_resource(
    bucket: str,
    *,
    prefix: str | None = None,
    suffix: str | None = None,
    key_type: type,
    value_type: type,
    compress: Any = None,
    encode_keys: bool = False,
    config: Config | None = None,
    policies: Iterable[Policy] | None = None,
):
    """
    Create S3 bucket resource.

    Parameters:
    • bucket: name of bucket to contain objects
    • prefix: prefix for all objects
    • suffix: suffix for all objects
    • key_type: type of key to identify object
    • value_type: type of value stored in each object
    • compress: algorithm to compress and decompress content
    • encode_keys: URL encode and decode object keys
    • config: client configuration
    • policies: security policies to apply to all operations
    """

    prefix = prefix or ""
    suffix = suffix or ""

    key_codec = fondat.codec.get_codec(String, key_type)
    value_codec = fondat.codec.get_codec(Binary, value_type)

    @asynccontextmanager
    async def create_client():
        async with fondat.aws.client.create_client("s3", config=config) as client:
            yield client

    @resource
    class Object:
        """S3 object resource."""

        def __init__(self, key: key_type):
            key = key_codec.encode(key)
            if encode_keys:
                key = quote(key, safe="")
            self._key = f"{prefix}{key}{suffix}"

        @operation(policies=policies)
        async def get(self) -> value_type:
            async with create_client() as client:
                try:
                    response = await client.get_object(Bucket=bucket, Key=self._key)
                    async with response["Body"] as stream:
                        body = await stream.read()
                    if compress:
                        body = compress.decompress(body)
                    return value_codec.decode(body)
                except client.exceptions.NoSuchKey:
                    raise NotFoundError
                except Exception as e:
                    _logger.error(e)
                    raise InternalServerError from e

        @operation(policies=policies)
        async def put(self, value: value_type) -> None:
            body = value_codec.encode(value)
            if compress:
                body = compress.compress(body)
            async with create_client() as client:
                try:
                    await client.put_object(Bucket=bucket, Key=self._key, Body=body)
                except Exception as e:
                    _logger.error(e)
                    raise InternalServerError from e

        @operation(policies=policies)
        async def delete(self) -> None:
            async with create_client() as client:
                try:
                    await client.delete_object(Bucket=bucket, Key=self._key)
                except Exception as e:
                    _logger.error(e)
                    raise InternalServerError from e

    Page = fondat.pagination.make_page_dataclass("Page", str)

    @resource
    class Bucket:
        """S3 bucket resource."""

        @operation(policies=policies)
        async def get(
            self,
            limit: int | None = None,
            cursor: bytes | None = None,
        ) -> Page:
            kwargs = {}
            if limit and limit > 0:
                kwargs["MaxKeys"] = limit
            if prefix:
                kwargs["Prefix"] = prefix
            if cursor is not None:
                kwargs["ContinuationToken"] = cursor.decode()
            async with create_client() as client:
                try:
                    response = await client.list_objects_v2(Bucket=bucket, **kwargs)
                    next_token = response.get("NextContinuationToken")
                    page = Page(
                        items=[
                            content["Key"][len(prefix) : len(content["Key"]) - len(suffix)]
                            for content in response.get("Contents", ())
                        ],
                        cursor=next_token.encode() if next_token else None,
                        remaining=None,
                    )
                    return page
                except Exception as e:
                    _logger.error(e)
                    raise InternalServerError from e

        def __getitem__(self, key: key_type) -> Object:
            return Object(key)

    return Bucket()



"""..."""

import fondat.sql

from collections.abc import AsyncIterator
from contextlib import asynccontextmanager
from fondat.sql import Expression
from typing import Any


class Database:
    """..."""

    pass


class Table:
    """..."""

    def __init__(
        database: Database,
        name: str,
        columns: None,
        partitioned_by: list[str],
        location: str,
        properties: dict[str, str],
    ):
        pass


def athena_database_resource(config: Any) -> Any:
    pass



"""Fondat module for AWS CloudWatch."""

import asyncio
import logging

from collections import deque
from collections.abc import Iterable
from copy import copy
from datetime import datetime, timezone
from fondat.aws.client import Config, create_client
from fondat.codec import JSON, get_codec
from fondat.data import datacls
from fondat.http import AsBody
from fondat.monitor import Measurement, Monitor
from fondat.resource import operation, resource
from fondat.security import Policy
from fondat.validation import validate_arguments
from typing import Annotated, Any, Literal


_logger = logging.getLogger(__name__)


Unit = Literal[
    "Seconds",
    "Microseconds",
    "Milliseconds",
    "Bytes",
    "Kilobytes",
    "Megabytes",
    "Gigabytes",
    "Terabytes",
    "Bits",
    "Kilobits",
    "Megabits",
    "Gigabits",
    "Terabits",
    "Percent",
    "Count",
    "Bytes/Second",
    "Kilobytes/Second",
    "Megabytes/Second",
    "Gigabytes/Second",
    "Terabytes/Second",
    "Bits/Second",
    "Kilobits/Second",
    "Megabits/Second",
    "Gigabits/Second",
    "Terabits/Second",
    "Count/Second",
    "None",
]


unit_conversions = {
    "s": "Seconds",
    "μs": "Microseconds",
    "ms": "Milliseconds",
    "B": "bytes",
    "kB": "Kilobytes",
    "MB": "Megabytes",
    "GB": "Gigabytes",
    "TB": "Terabytes",
    "b": "Bits",
    "kb": "Kilobits",
    "Mb": "Megabits",
    "Gb": "Gigabits",
    "Tb": "Terabits",
    "%": "Percent",
    "B/s": "Bytes/Second",
    "kB/s": "Kilobytes/Second",
    "MB/s": "Megabytes/Second",
    "GB/s": "Gigabytes/Second",
    "TB/s": "Terabytes/Second",
    "b/s": "Bits/Second",
    "kb/s": "Kilobits/Second",
    "Mb/s": "Megabits/Second",
    "Gb/s": "Gigabits/Second",
    "TB/s": "Terabits/Second",
}


@datacls
class Dimension:
    name: str
    value: str


@datacls
class StatisticSet:
    sample_count: float
    sum: float
    minimum: float
    maximum: float


@datacls
class MetricDatum:
    metric_name: str
    dimensions: list[Dimension] | None
    timestamp: datetime | None
    value: float | None
    statistic_values: StatisticSet | None
    values: list[float] | None
    counts: list[float] | None
    unit: Unit | None
    storage_resolution: Literal[1, 60] | None


def _naming(value: Any) -> Any:
    match value:
        case dict():
            return {k.title().replace("_", ""): _naming(v) for k, v in value.items()}
        case list():
            return [_naming(i) for i in value]
        case _:
            return value


def _awsify(value: Any) -> dict[str, Any]:
    return _naming(get_codec(JSON, type(value)).encode(value))


def cloudwatch_resource(
    *,
    config: Config | None = None,
    policies: Iterable[Policy] | None = None,
):
    """
    Create CloudWatch resource.

    Parameters:
    • config: client configuration
    • policies: security policies to enforce for resource operations
    """

    @resource
    class NamespaceResource:
        """TODO."""

        def __init__(self, name: str):
            self.name = name

        @operation(policies=policies)
        async def post(
            self,
            metric_data: Annotated[list[MetricDatum], AsBody, "metrics data to post"],
        ):
            """Post metrics to namespace."""

            async with create_client("cloudwatch", config=config) as client:
                await client.put_metric_data(
                    Namespace=self.name,
                    MetricData=_awsify(metric_data),
                )

    @resource
    class CloudWatchResource:
        """Create CloudWatch resource."""

        def namespace(self, name: str) -> NamespaceResource:
            """TODO."""
            return NamespaceResource(name)

    return CloudWatchResource()


def _ascii(value: str) -> str:
    return value.encode("ASCII", "xmlcharrefreplace").decode("ASCII")


class CloudWatchMonitor(Monitor):
    """
    Monitor that publishes recorded measurements in AWS CloudWatch.

    Parameters:
    • namespace: TODO
    • storage_resolution: TODO
    • cache_size: TODO
    • config: TODO
    """

    @validate_arguments
    def __init__(
        self,
        namespace: str,
        *,
        storage_resolution: Literal[1, 60] = 60,
        cache_size: int = 100,
        config: Config | None = None,
    ):
        self.storage_resolution = storage_resolution
        self.cache_size = cache_size
        self._namespace_resource = cloudwatch_resource(config=config).namespace(namespace)
        self._measurements = []
        self._task = None

    def _round(self, timestamp: datetime) -> datetime:
        ts = int(timestamp.timestamp())  # truncate milliseconds
        ts -= ts % self.storage_resolution  # round to beginning of resolution interval
        return datetime.fromtimestamp(ts, tz=timezone.utc)

    async def _flush(self):
        measurements = deque()
        for measurement in self._measurements:
            measurement = copy(measurement)
            measurement.timestamp = self._round(measurement.timestamp)
            measurement.value = float(measurement.value)
            measurements.append(measurement)
        self._measurements = []  # prevent race between record and flush
        metric_data = []
        while measurements:
            m1 = measurements.popleft()
            datum = MetricDatum(
                metric_name=_ascii(m1.name),
                dimensions={_ascii(k): _ascii(v) for k, v in m1.tags.items()},
                timestamp=m1.timestamp,
                statistic_values=StatisticSet(
                    sample_count=1.0,
                    sum=m1.value,
                    minimum=m1.value,
                    maximum=m1.value,
                ),
                unit=unit_conversions.get(m1.unit) or "Counter"
                if m1.type == "counter"
                else None,
                storage_resolution=self.storage_resolution,
            )
            n = 0
            while n < len(measurements):
                m2 = measurements[n]
                if m2.name == m1.name and m2.tags == m1.tags and m2.timestamp == m1.timestamp:
                    statistic_values = datum.statistic_values
                    statistic_values.sample_count += 1.0
                    statistic_values.sum += m2.value
                    statistic_values.minimum = min(statistic_values.minimum, m2.value)
                    statistic_values.maximum = max(statistic_values.maximum, m2.value)
                    del measurements[n]
                else:
                    n += 1
            metric_data.append(datum)
            if len(metric_data) == 20 or not measurements:
                try:
                    await self._namespace_resource.post(metric_data=metric_data)
                except Exception as e:
                    _logger.exception("failure posting metric data")
                metric_data = []

    async def record(self, measurement: Measurement):
        """Record a measurement."""
        if len(measurement.tags) > 10:
            raise ValueError("measurement tags exceeds limit of 10")
        self._measurements.append(measurement)
        if len(self._measurements) >= self.cache_size:
            await self.flush(block=False)

    async def flush(self, block: bool = True):
        """Flush all recorded measurements."""
        if self._task and self._task.done():
            self._task = None
        if not self._task and self._measurements:
            self._task = asyncio.create_task(self._flush())
        if block and self._task:
            await self._task



"""Fondat module for AWS Secrets Manager."""

import fondat.aws.client
import logging

from collections.abc import Iterable
from contextlib import asynccontextmanager, suppress
from fondat.aws.client import Config, wrap_client_error
from fondat.data import datacls
from fondat.error import NotFoundError
from fondat.http import AsBody, InBody
from fondat.memory import memory_resource
from fondat.resource import operation, resource
from fondat.security import Policy
from typing import Annotated, Any


_logger = logging.getLogger(__name__)


@datacls
class Secret:
    value: str | bytes


def secrets_resource(
    cache_size: int = 0,
    cache_expire: int | float = 1,
    config: Config | None = None,
    policies: Iterable[Policy] | None = None,
) -> Any:
    """
    Create secrets resource.

    Parameters:
    • cache: time in seconds to cache secrets
    • config: client configuration
    • policies: security policies to apply to all operations
    """

    @asynccontextmanager
    async def create_client():
        async with fondat.aws.client.create_client("secretsmanager", config=config) as client:
            yield client

    cache = (
        memory_resource(
            key_type=str,
            value_type=Secret,
            size=cache_size,
            evict=True,
            expire=cache_expire,
        )
        if cache_size
        else None
    )

    @resource
    class SecretResource:
        """..."""

        __slots__ = ("name",)

        def __init__(self, name: str):
            self.name = name

        @operation(policies=policies)
        async def get(
            self,
            version_id: str | None = None,
            version_stage: str | None = None,
        ) -> Secret:
            """Get secret."""
            if cache:
                with suppress(NotFoundError):
                    return await cache[self.name].get()
            kwargs = {}
            kwargs["SecretId"] = self.name
            if version_id is not None:
                kwargs["VersionId"] = version_id
            if version_stage is not None:
                kwargs["VersionStage"] = version_stage
            async with create_client() as client:
                with wrap_client_error():
                    value = await client.get_secret_value(**kwargs)
                secret = Secret(value=value.get("SecretString") or value.get("SecretBinary"))
                if cache:
                    await cache[self.name].put(secret)
                return secret

        @operation(policies=policies)
        async def put(self, secret: Annotated[Secret, AsBody]):
            """Update secret."""
            args = {
                "SecretString"
                if isinstance(secret.value, str)
                else "SecretBinary": secret.value
            }
            async with create_client() as client:
                with wrap_client_error():
                    await client.put_secret_value(SecretId=self.name, **args)
                if cache:
                    await cache[self.name].put(secret)

        @operation(policies=policies)
        async def delete(self):
            """Delete secret."""
            if cache:
                with suppress(NotFoundError):
                    await cache[self.name].delete()
            async with create_client() as client:
                with wrap_client_error():
                    await client.delete_secret(SecretId=self.name)

    @resource
    class SecretsResource:
        """..."""

        @operation(policies=policies)
        async def post(
            self,
            name: Annotated[str, InBody],
            secret: Annotated[Secret, InBody],
            kms_key_id: Annotated[str | None, InBody] = None,
            tags: Annotated[dict[str, str] | None, InBody] = None,
        ):
            """Create secret."""
            kwargs = {}
            kwargs["Name"] = name
            if isinstance(secret.value, str):
                kwargs["SecretString"] = secret.value
            else:
                kwargs["SecretBinary"] = secret.value
            if kms_key_id is not None:
                kwargs["KmsKeyId"] = kms_key_id
            if tags is not None:
                kwargs["Tags"] = [{"Key": k, "Value": v} for k, v in tags.items()]
            async with create_client() as client:
                with wrap_client_error():
                    await client.create_secret(**kwargs)
                if cache:
                    await cache[name].put(secret)

        def __getitem__(self, name: str) -> SecretResource:
            return SecretResource(name)

    return SecretsResource()
